@@ -27,12 +27,68 @@ Varian .xim FIle Format Parser
 import struct
 import numpy as np
 
-NP_BLUT = {1:np.int8, 2:np.int16, 4:np.int32}
-#S_BLUT = {1:'b', 2:'h', 4:'i'}
+try:
+  import numba
+except:
+  numba = None
 
 class EmptyObject(object):
+    '''Empty object for use in reading the histogram'''
+    # TODO: Consider creating a proper histogram object in the future as I refactor the pymp libraries
     pass
 
+# A BIG THANKS to Reese Haywood for the numpy array vs dict lookup suggestion
+NP_BLUT = np.array([(np.int8),(np.int16),(np.int32)])
+
+# Compile the LAME pixel by pixel build 3.7 python should allow the try to go in
+# 3.6 doesn't support it
+#
+# A BIG THANKS to Reese Haywood for the numba suggestion to speed up decompression
+if numba is not None: # If you have numba use it
+    try:
+        # The numba compiled pixel decompression for int16 arrays
+        @numba.jit((numba.int64,numba.int16[:],numba.int64))
+        def calc_pixel_values16(image_pos,pixel_data,image_width):
+            for i in pixel_data[image_width+1:]:
+                pixel_data[image_pos] = i + pixel_data[image_pos-1] + \
+                        pixel_data[image_pos-image_width] - pixel_data[image_pos-image_width-1]
+                image_pos += 1
+
+        # The numba compiled pixel decompression for int32 arrays
+        @numba.jit((numba.int64, numba.int32[:], numba.int64))
+        def calc_pixel_values32(image_pos, pixel_data, image_width):
+            for i in pixel_data[image_width + 1:]:
+                pixel_data[image_pos] = i + pixel_data[image_pos - 1] + \
+                                        pixel_data[image_pos - image_width] - pixel_data[image_pos - image_width - 1]
+                image_pos += 1
+    # If numba fails to compile let the user know and revert back to standard python
+    except Exception as e:
+        print(e)
+        numba = None
+
+
+class XIMHeader(object):
+    """ A Blank object to hold and neatly display the XIM header values after parsing.
+    """
+    def __str__(self):
+        """ The String Representation of the XIM header object
+        """
+        str_repr = f"""
+        XIM 32 byte fixed header:
+        ---------------------------------------------------------------------------
+        {'Element':30}{'Type':15}{'Value':10}
+        ---------------------------------------------------------------------------
+        {'File Format Identifier':30}{'Char1':15}{getattr(self, 'format_identifier', None)} 
+        {'File Format Version':30}{'Int4':15}{getattr(self, 'version', None)} 
+        {'Image Width in pixels':30}{'Int4':15}{getattr(self, 'image_width', None)} 
+        {'Image Height in pixels':30}{'Int4':15}{getattr(self, 'image_height', None)} 
+        {'Bits Per Pixel':30}{'Int4':15}{getattr(self, 'bits_per_pixel', None)} 
+        {'Bytes Per Pixel':30}{'Int4':15}{getattr(self, 'bytes_per_pixel', None)}
+        {'Compression indicator':30}{'Int4':15}{getattr(self, 'compression_indicator', None)}
+        ---------------------------------------------------------------------------
+        """
+        return str_repr
+
 class PropertyCollection(dict):
     """ Dict like collection of all the properties listed in a .xim file. Allows 
         dict like access  or dot (.) access to all the properties in the 
@@ -67,6 +123,11 @@ class PropertyCollection(dict):
         
         return dict(matches)
 
+    def __str__(self):
+        # return '\n'.join([f"{i:50}\t{self[i]['value']}" for i in self])
+
+        return '\n'.join([f"{i:50}{self[i]['value']}" for i in self])
+
 class VarianXIMImage(object):
     """ Simple file wrapper that gives you object like access to all the .xim
         file data.
@@ -121,15 +182,9 @@ class VarianXIMImage(object):
 
             lookup_table : numpy.ndarray
                 The numpy array of 2 bit compression flags indicating the number 
-                of bytes use to store each pixel differences used int he HND 
+                of bytes use to store each pixel differences used in the HND
                 compression (only set if compression is indicated in the header)
 
-            byte_table : nupy.ndarray
-                A table holding the actual number of bytes used to store the 
-                pixel differences at each corresponding position. Not part of 
-                the .xim format (only set if compression is indicated in the 
-                header).
-
             compressed_pixel_buffer_size : int
                 The size of the compressed pixel buffer in bytes (only set if 
                 compression is indicated in the header).
@@ -154,7 +209,7 @@ class VarianXIMImage(object):
             self.fileName = fileName
 
             # Add a blank 'object' to simply store the header information
-            self.header = EmptyObject()
+            self.header = XIMHeader()
             
             #------------------------------------------------------------------#
             # Section 1 - Header (32 bytes fixed size)                         #
@@ -275,13 +330,6 @@ class VarianXIMImage(object):
                 self.lookup_table[2::4] = np.right_shift(np.left_shift(lut,2),6)
                 self.lookup_table[3::4] = np.right_shift(np.left_shift(lut,0),6)
                 
-                # Quickly conver 2 bit flags into actual bytes per position
-                #
-                # NOTE: Not part of the xim format specification but much 
-                # faster than looking up the byte value for each flag in the 
-                # Lookup Table. 
-                self.byte_table = (2**self.lookup_table).astype(np.uint8)
-                
                 # Compressed Pixel Buffer Size (Type: Int4)
                 self.compressed_pixel_buffer_size, = struct.unpack('<i',
                                                                    f.read(4)
@@ -301,19 +349,27 @@ class VarianXIMImage(object):
                     # pixels for compression and the now decoded bytes required 
                     # for each difference stored located in the byte_table you 
                     # created
-                    
+
                     # SECTION 1: Uncompressed pixel values in first row and the 
                     # first pixel of second row are stored as int32s
-                    s1 = [('d{}'.format(i),np.int32) \
-                                    for i in range(self.header.image_width + 1)]
+                    s1 = [('',np.int32) for i in range(self.header.image_width + 1)]
                     
                     # SECTION 2: Differences are store in variable byte lengths
                     # as defined in the byte_table. Use a byte to numpy type 
                     # lookup dict to construct the numpy dtype entries
-                    s2 = [('d{}'.format(i+1281), 
-                           NP_BLUT[self.byte_table[i]]
-                           ) for i in range(len(self.byte_table)-1)]
-                    
+
+                    # Determine the unused lut values at the end of the table
+                    # If there are no unused values then the image (width * height) - (lookup_table_size * 4)
+                    # should equal the image width resulting in 'unused_lut_values' == 0
+                    unused_lut_values = self.header.image_width - \
+                                    ((self.header.image_width * self.header.image_height) - (self.lookup_table_size * 4))
+
+                    # Use only the usefull values in the table to generate the data types needed
+                    # NOTE: The (1 + unused_lut_values) in the indexing. The (1+) accounts for the first value of the
+                    #       second row of the images which also results in an unused lut value because of the redundancy
+                    np_types = NP_BLUT[self.lookup_table[0:-(1 + unused_lut_values)]]
+                    s2 = [('', val) for val in np_types]
+
                     # Combine the section into one composite numpy dtype
                     dt = np.dtype(s1 + s2)
                     
@@ -329,43 +385,53 @@ class VarianXIMImage(object):
                                                                count=1
                                                                )
                     
-                    # Conver the variable bit depth Compressed Pixel Buffer to
+                    # Convert the variable bit depth Compressed Pixel Buffer to
                     # the specified bit depth pixels (still only contains the 
                     # differences)
                     #
-                    # This step could actulally be seen as the decompression 
+                    # This step could actually be seen as the decompression
                     # step as represented in memory but the actual pixel values 
                     # still haven't been calculated
                     if self.header.bytes_per_pixel == 2:
                         self.pixel_data = \
-                            np.array(self.compressed_pixel_buffer.tolist()[0], 
+                            np.array(self.compressed_pixel_buffer.tolist()[0],
                                      dtype=np.int16
                                      )
                     else:
                         self.pixel_data = \
-                            np.array(self.compressed_pixel_buffer.tolist()[0], 
+                            np.array(self.compressed_pixel_buffer.tolist()[0],
                                      dtype=np.int32
                                      )
-                    
+
                     # Calculate the pixel values from the differences
                     #
-                    # NOTE: THIS IS WHERE YOU SPEND ALL YOUR TIME AS IT IS A 
+                    # NOTE: THIS IS WHERE YOU SPEND ALOT OF YOUR TIME AS IT IS A
                     # PIXEL-BY-PIXEL CALCULATION. LAME!!
                     #
+                    # BIG THANKS to Reese Haywood for the numba suggestion
+                    #
                     # NOTE: pixel_data = Uncompressed Pixel Buffer from the xim
                     # specification
                     image_pos = self.header.image_width+1
-                    for i in self.pixel_data[self.header.image_width+1:]:
-                        try:
-                            self.pixel_data[image_pos] = \
-                            i + self.pixel_data[image_pos-1] + \
-                            self.pixel_data[image_pos-self.header.image_width] - \
-                            self.pixel_data[image_pos-self.header.image_width-1]
+                    if numba is not None:
+                        # calc_pixel_values(image_pos, self.pixel_data, self.header.image_width)
+                        if self.pixel_data.dtype == np.dtype('int32'):
+                            calc_pixel_values32(image_pos, self.pixel_data, self.header.image_width)
+                        elif self.pixel_data.dtype == np.dtype('int16'):
+                            calc_pixel_values16(image_pos, self.pixel_data, self.header.image_width)
+                    else:
+                        for i in self.pixel_data[self.header.image_width+1:]:
+                            try:
+                                self.pixel_data[image_pos] = \
+                                i + self.pixel_data[image_pos-1] + \
+                                self.pixel_data[image_pos-self.header.image_width] - \
+                                self.pixel_data[image_pos-self.header.image_width-1]
+
+                                image_pos += 1
+                            except:
+                                raise \
+                                ValueError("Something went wrong in pixel calc")
 
-                            image_pos += 1
-                        except:
-                            raise \
-                            ValueError("Something went wrong in pixel calc")
                             
                     # Reshape the numpy array of pixels into a 2D image of
                     # (Image Height x Image Width) per numpy axis ordering
@@ -526,5 +592,4 @@ class VarianXIMImage(object):
                     setattr(self.properties, 
                             _property['name'], 
                             _property['value']
-                            )
-                        
\ No newline at end of file
+                            )
\ No newline at end of file
